-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "auxiliary"
M.relevantDevice = nil

local max = math.max
local random = math.random

local bombIndex = 1
local activeBombs = {}
local activeBombCount = 0
local availableBombs = {}

local switchedToFreeCam = false
local bombCamRefNodeData = nil
local ownID = obj:getID()

local function setCameraRefNodes(center, left, back)
  obj:queueGameEngineLua(string.format("core_camera.setRefNodes(%d, %d, %d, %d)", ownID, center, left, back))
end

local function resetCameraRefNodes()
  local refNodes = v.data.refNodes[0]
  setCameraRefNodes(refNodes.ref, refNodes.left, refNodes.back)
end

local function setCameraFree()
  obj:queueGameEngineLua("commands.setCameraFree()")
  switchedToFreeCam = true
end

local function setCameraPlayer()
  if switchedToFreeCam then
    obj:queueGameEngineLua("commands.setCameraPlayer()")
    switchedToFreeCam = false
  end
end

local function cancelWeaponCam()
  bombCamRefNodeData = nil
  resetCameraRefNodes()
  setCameraPlayer()
end

local function updateGFX(dt)
  local activeBombs2 = {}
  for _,v in ipairs(activeBombs) do
    if v.isAlive then
      if obj:beamIsBroken(v.collisionCid) then
        for _,cid in ipairs(v.warheadCids) do
          obj:setNodeMass(cid, v.explosionNodeMass)
        end
        sounds.playSoundOnceAtNode("missileExplodeSound", (v.nodeCid), 15)
        for _,n in ipairs(v.fireCids) do
          local random = random(2) - 1
          obj:addParticleByNodes(v.nodeCid, n, 50 * random, 31, 3, 100)
        end
        v.state = "dead"
        v.isAlive = false

        if bombCamRefNodeData and bombCamRefNodeData.bombID == v.id then
          bombCamRefNodeData = nil
          resetCameraRefNodes()
          if playerInfo.anyPlayerSeated then
            setCameraFree()
          end
        end
      end
      table.insert(activeBombs2, v)
    end
  end

  activeBombs = activeBombs2
  activeBombCount = #activeBombs

  if bombCamRefNodeData then
    bombCamRefNodeData.timer = max(bombCamRefNodeData.timer - dt, 0)
    if bombCamRefNodeData.timer <= 0 and not bombCamRefNodeData.hasSetRefNodes then
      setCameraRefNodes(bombCamRefNodeData.center, bombCamRefNodeData.left, bombCamRefNodeData.back)
      bombCamRefNodeData.hasSetRefNodes = true
    end
  end
end

local function deployBomb()
  if availableBombs[bombIndex] then
    local bomb = availableBombs[bombIndex]
    bomb.state = "dropping"
    bomb.isAlive = true
    bomb.hasFired = true
    table.insert(activeBombs, bomb)
    activeBombCount = #activeBombs
    bombIndex = bombIndex - 1
    beamstate.detachCouplers(bomb.couplerTag)

    bombCamRefNodeData = {timer = 0.5, bombID = bomb.id, center = bomb.frontCid, left = bomb.leftCid, back = bomb.backCid}
  end

  gui.message(string.format("Bombs: %d", max(bombIndex, 0)), 3, "bombs")
end

local function init(jbeamData)
  --dump(jbeamData)

  activeBombs = {}
  availableBombs = {}
  activeBombCount = 0

  bombCamRefNodeData = nil

  local bombExplosionNodeMass = jbeamData.explosionNodeMass or 150

  local bombFireRefs = {}
  local bombWarheadRefs = {}
  for k, node in pairs (v.data.nodes) do
    if node.fireNodeID then
      if not bombFireRefs[node.fireNodeID] then
        bombFireRefs[node.fireNodeID]= {}
      end
      table.insert(bombFireRefs[node.fireNodeID], k)
    end
    if node.warheadNodeID then
      if not bombWarheadRefs[node.warheadNodeID] then
        bombWarheadRefs[node.warheadNodeID]= {}
      end
      table.insert(bombWarheadRefs[node.warheadNodeID], k)
    end
  end

  --dump(bombFireRefs)

  local collisionDetectorRef = {}
  for _, beam in pairs(v.data.beams) do
    if beam.collisionDetectorID then collisionDetectorRef[beam.collisionDetectorID] = beam.cid end
  end

  for k,v in pairs(jbeamData) do
    if string.sub(k, 1, 4) == "bomb" then
      local bombID = v.id
      local bomb = {
        id = bombID,
        state = "idle",
        nodeCid = v.centerNode,
        frontCid = v.refFront,
        backCid = v.refBack,
        leftCid = v.refLeft,
        couplerTag = "bmrk_"..bombID,
        collisionCid = collisionDetectorRef[bombID],
        fireCids = bombFireRefs[bombID],
        warheadCids = bombWarheadRefs[bombID],
        explosionNodeMass = v.explosionNodeMass or bombExplosionNodeMass,
      }
      table.insert(availableBombs, bomb)
    end
  end

  table.sort(availableBombs,
    function(a,b)
      return a.id < b.id
    end)

  --dump(availableBombs)

  bombIndex = #availableBombs

  math.randomseed(os.time())
  resetCameraRefNodes()
end

M.init = init
M.updateGFX = updateGFX
M.deployWeapon = deployBomb
M.cancelWeaponCam = cancelWeaponCam
M.toggleTargetMode = nop

return M