-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "auxiliary"
M.relevantDevice = nil

local max = math.max
local autoCounterMeasure = false

local positionOffsetCoef = 0
local countermeasureActiveTime = 0
local countermeasureActiveTimer = 0
local countermeasureDelayTime = 0
local countermeasureDelayTimer = 0
local countermeasureActive = false
local emitterPositions = nil

local ownID = obj:getID()

local function activateCountermeasures()
  if countermeasureDelayTimer <= 0 and not countermeasureActive then
    countermeasureActiveTimer = countermeasureActiveTime
    countermeasureActive = true
  end
end

local function updateGFX(dt)
  electrics.values.missileWarning = 1

  local planePos = vec3(obj:getPosition())
  for k,v in pairs(mapmgr.objects or {}) do
    if k ~= ownID and tonumber(string.sub(k, 1, string.len(ownID))) ~= ownID then
      local targetVect = planePos - v.pos
      local targetVectLength = targetVect:length()
      local speedTowardsPlane = targetVect:dot(v.vel) / targetVectLength
      if speedTowardsPlane > 400 and targetVectLength < 2000 then
        if playerInfo.anyPlayerSeated then
          gui.message("Incoming missile!!!", 0.5, "missilewarning")
          electrics.values.missileWarning = 0
        end
        if autoCounterMeasure and targetVectLength < 800 then
          activateCountermeasures()
        end
      end
    end
  end

  if countermeasureActive then
    countermeasureActiveTimer = max(countermeasureActiveTimer - dt, 0)
    if countermeasureActiveTimer <= 0 then
      countermeasureActive = false
      countermeasureDelayTimer = countermeasureDelayTime
    end
    obj:queueGameEngineLua(string.format('map.objectData(%s,%s,%s,%s,%s,%s,%s,%s)',
        objectId,
        tostring(playerInfo.anyPlayerSeated),
        vec3(obj:getPosition()) - vec3(obj:getDirectionVector()) * positionOffsetCoef,
        vec3(obj:getVelocity()),
        vec3(obj:getDirectionVector()),
        vec3(obj:getDirectionVectorUp()),
        math.floor(beamstate.damage),
        serialize(beamstate.objectCollisionIds)))
    gui.message(string.format("Countermeasures are active for %.1fs", countermeasureActiveTimer), 0.5, "countermeasures")
    for _,p in ipairs(emitterPositions) do
      obj:addParticleByNodesRelative(p.front, p.back, 1, 6, 1, 10)
      obj:addParticleByNodesRelative(p.front, p.back, -1, 40, 1, 10)
    end
  end

  if countermeasureDelayTimer > 0 then
    countermeasureDelayTimer = max(countermeasureDelayTimer - dt, 0)
    if countermeasureDelayTimer > 0 then
      gui.message(string.format("Countermeasures are inactive for %.1fs", countermeasureDelayTimer), 0.5, "countermeasures")
    else
      gui.message("Countermeasures are ready", 3.5, "countermeasures")
    end
  end
end

local function init(jbeamData)
  electrics.values.missileWarning = 1

  positionOffsetCoef = jbeamData.positionOffsetCoef or 50
  countermeasureActiveTime = jbeamData.activeTime or 5
  countermeasureDelayTime = jbeamData.delayTime or 30

  autoCounterMeasure = jbeamData.autoCounterMeasure or false

  emitterPositions = {}
  local emitterPositionLookup = {}
  for k,e in pairs(jbeamData.emitters) do
    emitterPositionLookup[e[1]] = {location = "front", id = k}
    emitterPositionLookup[e[2]] = {location = "back", id = k}
    emitterPositions[k] = {}
  end

  for _,n in pairs(v.data.nodes) do
    if emitterPositionLookup[n.name] then
      local data = emitterPositionLookup[n.name]
      emitterPositions[data.id][data.location] = n.cid
    end
  end

  countermeasureActive = false
  countermeasureActiveTimer = 0
  countermeasureDelayTimer = 0
end

M.init = init
M.updateGFX = updateGFX
M.activateCountermeasures = activateCountermeasures

return M