-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "main"
M.relevantDevice = nil
M.engineInfo = {}
M.gearboxInfo = {isReverse = false}
M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0

M.weaponController = nil

local min = math.min --pull math.min/max to local scope for faster access (and less text)
local max = math.max
local abs = math.abs

local engineNodes = {L = {}, R = {}}
local barrelNodes = {front = {}, back= {}}
local leftEngine = nil
local rightEngine = nil

local gearTimer = 0

local energyStorageData = {
  tanks = {},
  ratio = 0,
  volume = 0,
  capacity = 0,
  invEnergyStorageCount = 0,
}

local bulletBeams = {}
local bulletNodes = {}
local activeBullets = {}
local activeBulletCount = 0
local recoilCounter = 0
local muzzleFlashCounter = 0
local rofTimer = 0
local numberOfRounds = nil
local roundIndex = nil
local fireRateTime = nil
local bulletTargetSpeed = 0
local realBulletSpeed = 540
local realBulletSpeedSquared = realBulletSpeed * realBulletSpeed
local quarterBulletTargetSpeed = 0
local additionalExplosiveEnergy = 0 --J

local leftWheel = nil
local rightWheel = nil

local flapsOff = 0
local flapsBattle = 0.2
local flapsTakeoff = 0.4
local flapsLanding = 1

local function toggleFlaps()
  if electrics.values.flap > 0 then
    electrics.values.flap = flapsOff
  elseif electrics.values.gear == 0 and electrics.values.airspeed <= 60 then
    electrics.values.flap = flapsTakeoff
    gui.message("Takeoff flaps", 2, "flapmessage")
  elseif electrics.values.gear == 0 and electrics.values.airspeed > 60 then
    electrics.values.flap = flapsLanding
    gui.message("Landing flaps", 2, "flapmessage")
  elseif electrics.values.gear ~= 0 then
    electrics.values.flap = flapsBattle
    gui.message("Combat flaps", 2, "flapmessage")
  end
end

local function deployWeapon()
  if not M.weaponController then
    return
  end

  M.weaponController.deployWeapon()
end

local function cancelWeaponCam()
  if not M.weaponController then
    return
  end

  M.weaponController.cancelWeaponCam()
end

local function toggleTargetMode()
  if not M.weaponController then
    return
  end

  M.weaponController.toggleTargetMode()
end

local function updateBulletFiring(bullet, dt)
  local nodeVelocity = obj:getNodeVelocityVector(bullet.nodeCid):length() - bullet.velocityOffset
  if nodeVelocity >= bulletTargetSpeed then
    electrics.values[bullet.thrusterSource] = nil
    recoilCounter = recoilCounter - 1
    local nodeMass = obj:getNodeMass(bullet.nodeCid)
    local targetEnergy = 0.5 * nodeMass * realBulletSpeedSquared + additionalExplosiveEnergy
    local equivalentWeight = min(max(targetEnergy * 2 / (nodeVelocity * nodeVelocity), nodeMass), 100)
    obj:setNodeMass(bullet.nodeCid, equivalentWeight)
    bullet.roundState = "fired"
    bullet.update = nop
  end
end

local function update(dt)
  for i = 1, activeBulletCount, 1 do
    activeBullets[i]:update(dt)
  end

  electrics.values.recoil = recoilCounter > 0 and 1 or nil --activates the recoil thruster
end

local function updateGFX(dt)
  --------DIFFERENTIAL BRAKES--------
  --add up the general brake with the steering brakes
  local rudderBrakeInput = electrics.values.airspeed < 30 and (input.rudder or 0) * 0.5 or 0
  local brakeR = min(max(electrics.values.brake_combined + electrics.values.brake_r + rudderBrakeInput, 0), 1)
  local brakeL = min(max(electrics.values.brake_combined + electrics.values.brake_l - rudderBrakeInput, 0), 1)

  --scale the parking brake torque to match our desired torque
  rightWheel.desiredBrakingTorque = rightWheel.brakeTorque * brakeR
  leftWheel.desiredBrakingTorque = leftWheel.brakeTorque * brakeL


  --------THRUST--------
  electrics.values.jet_throttle = min(max(electrics.values.jet_throttle + electrics.values.jet_throttle_change * dt * 0.5, 0), 1) --increase/decrease throttle via the button bindings

  local thrustLeft = leftEngine and leftEngine.throttleSmoothed or 0
  local thrustRight = rightEngine and rightEngine.throttleSmoothed or 0

  if thrustLeft > 0.15 then
    obj:addParticleByNodesRelative(engineNodes.L[1], engineNodes.L[2], 15, 40, 0.1, 1)
    if thrustLeft > 0.55 then
      obj:addParticleByNodesRelative(engineNodes.L[1], engineNodes.L[2], 25, 40, 0.1, 1)
    end
  end

  if thrustRight > 0.15 then
    obj:addParticleByNodesRelative(engineNodes.R[1], engineNodes.R[2], 15, 40, 0.1, 1)
    if thrustRight > 0.55 then
      obj:addParticleByNodesRelative(engineNodes.R[1], engineNodes.R[2], 25, 40, 0.1, 1)
    end
  end

  local engineLeftBroken = (leftEngine and leftEngine.isDisabled) or not leftEngine
  local engineRightBroken = (rightEngine and rightEngine.isDisabled) or not rightEngine

  electrics.values.engSoundL = engineLeftBroken and 1 or min(max(thrustLeft, 0.05), 0.99)
  electrics.values.engSoundR = engineRightBroken and 1 or min(max(thrustRight, 0.05), 0.99)

  electrics.values.exhaustL_low = engineLeftBroken and 0 or (thrustLeft < 0.25 and 0.5 or (thrustLeft < 0.5 and 1 or 0))
  electrics.values.exhaustL_high = engineLeftBroken and 0 or (thrustLeft < 0.5 and 0 or (thrustLeft > 0.9 and 1 or 0.5))
  electrics.values.exhaustR_low = engineRightBroken and 0 or (thrustRight < 0.25 and 0.5 or (thrustRight < 0.5 and 1 or 0))
  electrics.values.exhaustR_high = engineRightBroken and 0 or (thrustRight < 0.5 and 0 or (thrustRight > 0.9 and 1 or 0.5))

  --------ROCKET BOOSTERS--------
  if leftEngine then
    leftEngine:setBoosterState(thrustLeft >= 1 and leftEngine.hasBooster)
    electrics.values.boosterL = leftEngine.isBoosterActive
    electrics.values.boosterLSound = 1 - (electrics.values.boosterL and 1 or 0)
  end
  if rightEngine then
    rightEngine:setBoosterState(thrustRight >= 1 and rightEngine.hasBooster)
    electrics.values.boosterR = rightEngine.isBoosterActive
    electrics.values.boosterRSound = 1 - (electrics.values.boosterR and 1 or 0)
  end

  --------GUNS--------
  local activeBullets2 = {}
  for _,v in ipairs(activeBullets) do
    v.muzzleFlashTimer = max(v.muzzleFlashTimer - dt, 0)
    if v.muzzleFlashTimer <= 0 and v.hasMuzzleFlash then
      muzzleFlashCounter = muzzleFlashCounter - 1
      v.hasMuzzleFlash = false
    end
    if v.roundState == "fired" then
      local bulletVelocity = abs(obj:getNodeVelocityVector(v.nodeCid):length())
      if bulletVelocity <= quarterBulletTargetSpeed then
        obj:breakBeam(v.beamCid or 0)
        obj:addParticleByNodes(v.nodeCid, 0.1, 0, 31, 0, 1)
        if v.hasMuzzleFlash then
          muzzleFlashCounter = muzzleFlashCounter - 1
          v.hasMuzzleFlash = false
        end
        v.roundState = "dead"
      end
    end
    if v.roundState ~= "dead" then
      table.insert(activeBullets2, v)
    end
  end

  activeBullets = activeBullets2
  --dump(activeBullets)

  rofTimer = max(rofTimer - dt, 0)
  if electrics.values.guns >= 1 and rofTimer <= 0 and roundIndex < numberOfRounds then --only go in here if we are STILL firing or want to fire
    sounds.playSoundOnceAtNode("autocannon", barrelNodes.front[1], 5)	--Play the firing sound
    sounds.playSoundOnceAtNode("autocannon", barrelNodes.front[2], 5)
    sounds.playSoundOnceAtNode("autocannon", barrelNodes.front[3], 5)
    sounds.playSoundOnceAtNode("autocannon", barrelNodes.front[4], 5)

    obj:addParticleByNodesRelative(barrelNodes.front[1], barrelNodes.back[1], 10, 6, 0, 1)
    obj:addParticleByNodesRelative(barrelNodes.front[2], barrelNodes.back[2], 10, 6, 0, 1)
    obj:addParticleByNodesRelative(barrelNodes.front[3], barrelNodes.back[3], 10, 6, 0, 1)
    obj:addParticleByNodesRelative(barrelNodes.front[4], barrelNodes.back[4], 10, 6, 0, 1)

    gui.message(string.format("Cannon rounds: %d", (numberOfRounds - roundIndex - 1) * 4), 3, "cannonrounds")
    for i = 1, 4, 1 do
      local thrusterSource = "shell_"..(roundIndex * 4 + i)
      electrics.values[thrusterSource] = 1 --roundIndex starts at 1, and determines which bullet is to be fired. it is added to "shell_", so if roundIndex is 1 then the thruster with "bllt_1" as the control is activated
      table.insert(activeBullets,
        {
          beamCid = bulletBeams[roundIndex * 4 + i],
          nodeCid = bulletNodes[roundIndex * 4 + i],
          thrusterSource = thrusterSource,
          velocityOffset = electrics.values.airspeed,
          muzzleFlashTimer = 0.05,
          hasMuzzleFlash = true,
          roundState = "firing",
          update = updateBulletFiring})
    end
    rofTimer = fireRateTime
    roundIndex = roundIndex + 1
    recoilCounter = recoilCounter + 4
    muzzleFlashCounter = muzzleFlashCounter + 4
  end

  activeBulletCount = #activeBullets
  --print(activeBulletCount)
  --dump(activeBullets)
  electrics.values.muzf = muzzleFlashCounter > 0 and 1 or 0	--show the muzzleflash. This is done with glowmaps with "muzf" as their input.


  --------LANDING GEAR DOORS--------
  if electrics.values.gear == 0 then    --if gear is out...
    gearTimer = 0                  --reset the gear timer...
    electrics.values.geardoors = 1    --and open the gear doors
  elseif electrics.values.geardoors ~= 0 then --only enter here if the geardoors aren't closed (yet)
    gearTimer = gearTimer + dt                --start timer
    electrics.values.brake_combined = 1	--apply brakes so wheels dont spin in their wells
    if gearTimer > 2.8 then            --trigger after geartimer reaches 2.8 seconds
      electrics.values.geardoors = 0  --close doors
      gearTimer = 0                  --reset timer
      electrics.values.brake_combined = 0 -- release brake
    end
  end

  --------INSTRUMENTS--------
  local dirVector = obj:getDirectionVector()
  local dirVectorUp = obj:getDirectionVectorUp()
  --local position = obj:getPosition()
  local inst_pitch = dirVector.z
  local inst_roll = dirVectorUp.x * -dirVector.y + dirVectorUp.y * dirVector.x
  --local yaw = obj:getDirection(),

  electrics.values.inst_pitch = inst_pitch
  electrics.values.inst_roll = inst_roll

  energyStorageData.ratio = 0
  energyStorageData.volume = 0

  for _,s in pairs(energyStorageData.tanks) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.ratio = energyStorageData.ratio + energyStorage.remainingRatio
      energyStorageData.volume = energyStorageData.volume + energyStorage.remainingVolume
    end
  end
  energyStorageData.ratio = energyStorageData.ratio * energyStorageData.invEnergyStorageCount

  electrics.values.fuel = energyStorageData.ratio or 0
  electrics.values.lowfuel = electrics.values.fuel < 0.1
  electrics.values.fuelCapacity = energyStorageData.invEnergyStorageCount > 0 and (energyStorageData.capacity or 0) or 1
  electrics.values.fuelVolume = energyStorageData.volume

  electrics.values.throttle = electrics.values.jet_throttle
  electrics.values.brake = electrics.values.brake_combined
  electrics.values.clutch = 0
  electrics.values.clutchRatio = 1
  electrics.values.gearIndex = 0
  electrics.values.rpm = max(thrustLeft, thrustRight) * 1000
  electrics.values.wheelspeed = electrics.values.airspeed

  M.throttle = electrics.values.jet_throttle
  M.brake = electrics.values.brake_combined

  M.engineInfo = {
    0
    , 1000
    , 0
    , 0
    , 0
    , string.format("%d%%", electrics.values.jet_throttle * 100)
    , 0
    , 0
    , 0
    , 0
    , obj:getVelocity():length()  -- airspeed
    , electrics.values.fuelVolume
    , electrics.values.fuelCapacity
    , "manual"
    , obj:getID()
    , 0
    , 0
    , 1
    , 1
  }
end

local function init(jbeamData)
  electrics.values.gear = 0 --gear input
  electrics.values.geardoors = 0 --gear door input
  electrics.values.jet_throttle = 0 --throttle
  electrics.values.jet_throttle_change = 0 --throttle
  electrics.values.guns = 0 --gun trigger
  electrics.values.recoil = nil --gun recoil
  electrics.values.muzf = 0 --muzzle flash
  electrics.values.brake_l = 0 --left wheel brake
  electrics.values.brake_r = 0	--right wheel brake
  electrics.values.brake_combined = 0
  electrics.values.open_canopy = 0

  electrics.values.engSoundL = 0
  electrics.values.engSoundR = 0

  electrics.values.boosterL = 0
  electrics.values.boosterR = 0
  electrics.values.boosterLSound = 0
  electrics.values.boosterRSound = 0

  electrics.values.exhaustL_low = 0
  electrics.values.exhaustL_high = 0
  electrics.values.exhaustR_low = 0
  electrics.values.exhaustR_high = 0

  --disable sounds for (potentially) non existing controllers
  electrics.values.lockingTone = 1
  electrics.values.lockTone = 1
  electrics.values.missileWarning = 1

  electrics.values.watertemp = 0

  gearTimer = 0 --Gear timer

  recoilCounter = 0
  muzzleFlashCounter = 0
  roundIndex = 0
  numberOfRounds = jbeamData.numberOfRounds or 0
  fireRateTime = 0.115
  bulletTargetSpeed = jbeamData.bulletSpeed or 410
  quarterBulletTargetSpeed = bulletTargetSpeed * 0.25
  additionalExplosiveEnergy = jbeamData.bulletAdditionalExplosiveEnergy or 536000

  engineNodes = {L= {}, R = {}}
  barrelNodes = {front = {}, back= {}}

  for k, node in pairs (v.data.nodes) do
    if node.name == "eng_l_62" then engineNodes.L[2] = k end
    if node.name == "eng_l_49" then engineNodes.L[1] = k end
    if node.name == "eng_r_62" then engineNodes.R[2] = k end
    if node.name == "eng_r_49" then engineNodes.R[1] = k end

    if node.name == "barrel_1_f" then barrelNodes.front[1] = k end
    if node.name == "barrel_2_f" then barrelNodes.front[2] = k end
    if node.name == "barrel_3_f" then barrelNodes.front[3] = k end
    if node.name == "barrel_4_f" then barrelNodes.front[4] = k end
    if node.name == "barrel_1_b" then barrelNodes.back[1] = k end
    if node.name == "barrel_2_b" then barrelNodes.back[2] = k end
    if node.name == "barrel_3_b" then barrelNodes.back[3] = k end
    if node.name == "barrel_4_b" then barrelNodes.back[4] = k end

    if node.bullet then bulletNodes[node.bullet] = node.cid  end
  end

  activeBullets = {}
  activeBulletCount = 0
  bulletBeams = {}

  for _, beam in pairs(v.data.beams) do
    if beam.tracer then bulletBeams[beam.tracer] = beam.cid  end
  end

  for _,wd in pairs(wheels.wheels) do
    if wd.name == "RW" then rightWheel = wd end
    if wd.name == "LW" then leftWheel = wd end
  end

  leftEngine = powertrain.getDevice("jetL")
  rightEngine = powertrain.getDevice("jetR")

  energyStorageData = {
    tanks = {},
    ratio = 0,
    volume = 0,
    capacity = 0,
    invEnergyStorageCount = 0,
  }

  local tankNames = {}

  if leftEngine then
    for _,s in pairs(leftEngine.energyStorage or {}) do
      tankNames[s] = s
    end
  end
  if rightEngine then
    for _,s in pairs(rightEngine.energyStorage or {}) do
      tankNames[s] = s
    end
  end

  for _,v in pairs(tankNames) do
    table.insert(energyStorageData.tanks, v)
  end

  local energyStorageCount = 0
  for _,s in ipairs(energyStorageData.tanks) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.capacity = energyStorageData.capacity + energyStorage.capacity
      energyStorageCount = energyStorageCount + 1
    end
  end

  energyStorageData.invEnergyStorageCount = energyStorageCount > 0 and 1 / energyStorageCount or 0

  math.randomseed(os.time())
  mapmgr.enableTracking()
  obj:queueGameEngineLua(string.format("be:getObjectByID(%d):setField('renderDistance', '', '%d')", obj:getID(), 15000))
end

local function initSecondStage()
  M.weaponController = controller.getController("missiles") or controller.getController("rockets") or controller.getController("bombs")
end

M.init = init
M.initSecondStage = initSecondStage
M.updateGFX = updateGFX
M.update = update
M.toggleFlaps = toggleFlaps
M.deployWeapon = deployWeapon
M.cancelWeaponCam = cancelWeaponCam
M.toggleTargetMode = toggleTargetMode

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setGearboxMode = nop
M.setStarter = nop
M.setFreeze = nop
M.sendTorqueData = nop
M.vehicleActivated = nop
-------------------------------

return M