-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "auxiliary"
M.relevantDevice = nil

local min = math.min --pull math functions to local scope for faster access (and less text)
local max = math.max
local abs = math.abs
local acos = math.acos
local random = math.random

local missileIndex = 1
local activeMissiles = {}
local activeMissileCount = 0
local availableMissiles = {}

local lockAngleVertical = nil
local lockAngleHorizontal = nil
local lockDistance = nil
local missileLockData = {}
local targetModes = {"Auto-Select", "Guided", "Freefire"}
local targetMode = targetModes[1]
local ownID = obj:getID()

local switchedToFreeCam = false
local missileCamRefNodeData = nil

local debugVectors = {}

local lockedTargets = {}

local function debugDraw(focusPos)
--  for _,v in pairs(debugVectors) do
--    obj.debugDrawProxy:drawNodeVector3d(v.thickness or 0.05, v.cid, (v.vector):toFloat3(), v.color)
--  end
  if playerInfo.anyPlayerSeated then
    local alreadyLockedIDs = {}
    local colorStep = 255 / #lockedTargets
    for k,v in pairs(lockedTargets) do
      obj.debugDrawProxy:drawSphere(v.distance / 35, v.position:toFloat3(), color((#lockedTargets - (k-1)) * colorStep,50,0,80))
      alreadyLockedIDs[v.id] = true
    end

    for k,v in pairs(mapmgr.objects or {}) do
      if k ~= ownID and tonumber(string.sub(k, 1, string.len(ownID))) ~= ownID and not alreadyLockedIDs[k] then
        local distance = (v.pos - vec3(obj:getPosition())):length()
        obj.debugDrawProxy:drawSphere(distance / 35, v.pos:toFloat3(), color(50,255,0,80))
      end
    end
  end
end

local function setCameraRefNodes(center, left, back)
  obj:queueGameEngineLua(string.format("core_camera.setRefNodes(%d, %d, %d, %d)", ownID, center, left, back))
end

local function resetCameraRefNodes()
  local refNodes = v.data.refNodes[0]
  setCameraRefNodes(refNodes.ref, refNodes.left, refNodes.back)
end

local function setCameraFree()
  obj:queueGameEngineLua("commands.setCameraFree()")
  switchedToFreeCam = true
end

local function setCameraPlayer()
  if switchedToFreeCam then
    obj:queueGameEngineLua("commands.setCameraPlayer()")
    switchedToFreeCam = false
  end
end

local function cancelWeaponCam()
  missileCamRefNodeData = nil
  resetCameraRefNodes()
  setCameraPlayer()
end

local function updateMissileExploding(missile, dt)
  for _,cid in ipairs(missile.warheadCids) do
    obj:setNodeMass(cid, missile.explosionNodeMass)
  end
  obj:breakBeam(missile.triggerCid)
  sounds.playSoundOnceAtNode("missileExplodeSound", (missile.thrusterNode2), 15)
  for _,n in ipairs(missile.fireCids) do
    local random = random(2) - 1
    obj:addParticleByNodes(missile.nodeCid, n, 50 * random, 31, 3, 100)
  end
  electrics.values[missile.motor] = nil
  electrics.values[missile.elevator] = 0
  electrics.values[missile.rudder] = 0
  missile.update = nop
  missile.state = "dead"
  missile.isAlive = false
  if missile.targetID and missileLockData[missile.targetID] then
    missileLockData[missile.targetID].activeMissiles = missileLockData[missile.targetID].activeMissiles - 1
  end
  if missileCamRefNodeData and missileCamRefNodeData.missileID == missile.id then
    missileCamRefNodeData = nil
    resetCameraRefNodes()
    if playerInfo.anyPlayerSeated then
      setCameraFree()
    end
  end
end

local function updateMissileFlyingGuided(missile, dt)
  local velocityVector = vec3(obj:getNodeVelocityVector(missile.nodeCid))
  local velocity = velocityVector:length()
  local velocityError = missile.targetVelocity - velocity
  local thrust = min(max(velocityError * 0.1, 0), 1)
  electrics.values[missile.motor] = thrust

  local target = mapmgr.objects[missile.targetID]
  if target then
    local posRefFront = vec3(obj:getNodePosition(missile.frontCid))
    local posRefBack = vec3(obj:getNodePosition(missile.backCid))
    local posRefLeft = vec3(obj:getNodePosition(missile.leftCid))
    local missilePos = vec3(obj:getPosition()) + vec3(obj:getNodePosition(missile.nodeCid))
    local targetPos = target.pos
    local targetVelocity = target.vel
    local distanceToTarget = (targetPos - missilePos):length()

    local aimAheadCoef = min(max(((missile.aimAheadCoefRange * (distanceToTarget - missile.aimAheadMinDistance)) * missile.aimAheadInvDistanceRange) + missile.aimAheadMinCoef, missile.aimAheadMinCoef), missile.aimAheadMaxCoef)
    local targetVector = (targetPos + targetVelocity * aimAheadCoef) - missilePos

    local directionVector = (posRefFront - posRefBack):normalized()
    local normalVertical = (posRefLeft - posRefFront):cross(posRefBack - posRefFront):normalized()
    local normalHorizontal = normalVertical:cross((posRefFront - posRefBack)):normalized()

    local projectedVertical = (targetVector - targetVector:dot(normalVertical) * normalVertical):normalized()
    local projectHorizontal = (targetVector - targetVector:dot(normalHorizontal) * normalHorizontal):normalized()

    local angleHorizontal = acos(directionVector:dot(projectedVertical))
    local horizontalAngleSign = fsign(normalVertical:dot(directionVector:cross(projectedVertical)))
    local angleVertical = acos(directionVector:dot(projectHorizontal))
    local verticalAngleSign = fsign(normalHorizontal:dot(directionVector:cross(projectHorizontal)))


    missile.position = missilePos
    missile.velocity = velocityVector
    missile.directionVector = directionVector
    missile.directionVectorUp = normalHorizontal

    electrics.values[missile.elevator] = min(max(verticalAngleSign * angleVertical * 2, -1), 1)
    electrics.values[missile.rudder] = min(max(horizontalAngleSign * angleHorizontal * 2, -1), 1)
    electrics.values[missile.elevatorVect] = min(max(verticalAngleSign * angleVertical * 2, -1), 1)
    electrics.values[missile.rudderVect] = min(max(horizontalAngleSign * angleHorizontal * 2, -1), 1)
--    debugVectors = {}

--    table.insert(debugVectors, {cid = missile.nodeCid, vector = normalVertical, color = color(255,0,0,255)})
--    table.insert(debugVectors, {cid = missile.nodeCid, vector = normalHorizontal, color = color(0,255,0,255)})
--    table.insert(debugVectors, {cid = missile.nodeCid, vector = directionVector * 10, color = color(128,128,0,255)})
--    table.insert(debugVectors, {cid = missile.nodeCid, vector = projectedVertical * 10, color = color(0,128,0,255)})
--    table.insert(debugVectors, {cid = missile.nodeCid, vector = projectHorizontal * 10, color = color(128,0,0,255)})
--    table.insert(debugVectors, {cid = missile.nodeCid, vector = targetVector, color = color(0,0,255,255), thickness = 0.1})

    --local thrusterVector = vec3(obj:getNodePosition(missile.thrusterNode2)) - vec3(obj:getNodePosition(missile.thrusterNode1))
    --table.insert(debugVectors, {cid = missile.thrusterNode2, vector = thrusterVector * 10, color = color(0,0,255,255), thickness = 0.1})

    if distanceToTarget < 10 then
      missile.state = "exploding"
      missile.update = updateMissileExploding
    end
  end
end

local function updateMissileFlyingNonGuided(missile, dt)
  local velocityVector = vec3(obj:getNodeVelocityVector(missile.nodeCid))
  local velocity = velocityVector:length()
  local velocityError = missile.targetVelocity - velocity
  local thrust = min(max(velocityError * 0.1, 0), 1)
  electrics.values[missile.motor] = thrust

  local posRefFront = vec3(obj:getNodePosition(missile.frontCid))
  local posRefBack = vec3(obj:getNodePosition(missile.backCid))
  local posRefLeft = vec3(obj:getNodePosition(missile.leftCid))

  local targetVector = missile.targetVector
  local directionVector = (posRefFront - posRefBack):normalized()
  local normalVertical = (posRefLeft - posRefFront):cross(posRefBack - posRefFront):normalized()
  local normalHorizontal = normalVertical:cross((posRefFront - posRefBack)):normalized()

  local projectedVertical = (targetVector - (targetVector:dot(normalVertical) / normalVertical:squaredLength()) * normalVertical):normalized()
  local projectHorizontal = (targetVector - (targetVector:dot(normalHorizontal) / normalHorizontal:squaredLength()) * normalHorizontal):normalized()

  local angleHorizontal = acos(directionVector:dot(projectedVertical))
  local horizontalAngleSign = fsign(normalVertical:dot(directionVector:cross(projectedVertical)))
  local angleVertical = acos(directionVector:dot(projectHorizontal))
  local verticalAngleSign = fsign(normalHorizontal:dot(directionVector:cross(projectHorizontal)))

  missile.position = vec3(obj:getPosition()) + vec3(obj:getNodePosition(missile.nodeCid))
  missile.velocity = velocityVector
  missile.directionVector = directionVector
  missile.directionVectorUp = normalHorizontal

  --debugVectors = {}

--  table.insert(debugVectors, {cid = missile.nodeCid, vector = normalVertical, color = color(255,0,0,255)})
--  table.insert(debugVectors, {cid = missile.nodeCid, vector = normalHorizontal, color = color(0,255,0,255)})
  --local thrusterVector = vec3(obj:getNodePosition(missile.thrusterNode2)) - vec3(obj:getNodePosition(missile.thrusterNode1))
  --table.insert(debugVectors, {cid = missile.thrusterNode2, vector = thrusterVector * 10, color = color(0,0,255,255), thickness = 0.1})

  electrics.values[missile.elevator] = min(max(verticalAngleSign * angleVertical * 2, -1), 1)
  electrics.values[missile.rudder] = min(max(horizontalAngleSign * angleHorizontal * 2, -1), 1)
  electrics.values[missile.elevatorVect] = min(max(verticalAngleSign * angleVertical * 2, -1), 1)
  electrics.values[missile.rudderVect] = min(max(horizontalAngleSign * angleHorizontal * 2, -1), 1)
end

local function updateMissileFiring(missile, dt)
  missile.timer = missile.timer + dt
  if missile.timer > 0.2 then
    missile.update = missile.mode == "guided" and updateMissileFlyingGuided or updateMissileFlyingNonGuided
    missile.state = "flying"
    missile.timer = 0
  end
end

local function update(dt)
  for i = 1, activeMissileCount, 1 do
    activeMissiles[i]:update(dt)
  end
end

local function updateGFX(dt)
  lockedTargets = {}
  local hasTarget = false
  local hasLockedTarget = false
  if availableMissiles[missileIndex] and targetMode ~= "Freefire" then
    local directionVector = vec3(obj:getDirectionVector())
    local position = vec3(obj:getPosition())
    for k,v in pairs(mapmgr.objects or {}) do
      if k ~= ownID and tonumber(string.sub(k, 1, string.len(ownID))) ~= ownID then
        local targetPos = v.pos
        local targetVector = targetPos - position
        local targetDistance = targetVector:length()
        local normalVertical = vec3(obj:getDirectionVectorUp()):normalized()
        local normalHorizontal = normalVertical:cross(directionVector):normalized()

        local projectedVertical = (targetVector - (targetVector:dot(normalVertical) / normalVertical:squaredLength()) * normalVertical):normalized()
        local projectHorizontal = (targetVector - (targetVector:dot(normalHorizontal) / normalHorizontal:squaredLength()) * normalHorizontal):normalized()

        local angleHorizontal = abs(acos(directionVector:dot(projectedVertical)))
        local angleVertical = abs(acos(directionVector:dot(projectHorizontal)))
        if not missileLockData[v.id] then
          missileLockData[v.id] = {lockTimer = 0, activeMissiles = 0}
        end
        if angleHorizontal < lockAngleHorizontal and angleVertical < lockAngleVertical and targetDistance <= lockDistance then
          missileLockData[v.id].lockTimer = missileLockData[v.id].lockTimer + dt
          hasTarget = true
          if missileLockData[v.id].lockTimer >= 2 then
            table.insert(lockedTargets, {id = v.id, lockScore = angleHorizontal * 100 + angleVertical * 100 + targetDistance + v.damage * 100 + missileLockData[v.id].activeMissiles * 1000000, position = targetPos, distance = targetDistance})
            hasLockedTarget = true
          end
        else
          missileLockData[v.id].lockTimer = 0
        end
      end
    end

    table.sort(lockedTargets, function(a,b) return a.lockScore < b.lockScore end)
    --dump(lockedTargets)
  else
    missileLockData = {}
  end

  electrics.values.lockTone = 1
  electrics.values.lockingTone = 1

  if playerInfo.anyPlayerSeated then
    if hasLockedTarget then
      gui.message("Missile has a lock!", 0.5, "missilelock")
      electrics.values.lockTone = 0
    elseif hasTarget then
      gui.message("Missile is locking...", 0.5, "missilelock")
      electrics.values.lockingTone = 0
    end
  end

  local activeMissiles2 = {}
  for _,v in ipairs(activeMissiles) do
    if v.isAlive then
      obj:queueGameEngineLua(string.format('map.objectData(%s,%s,%s,%s,%s,%s,%s,%s)', tostring(objectId)..v.id, false,
          v.position, v.velocity, v.directionVector, v.directionVectorUp, 0, "{}"))

      if obj:beamIsBroken(v.collisionCid) then
        v.update = updateMissileExploding
        v.state = "exploding"
      end

      v.fuel = max(v.fuel - dt, 0)
      if v.fuel <= 0 then
        v.update = updateMissileExploding
        v.state = "exploding"
      end
      table.insert(activeMissiles2, v)

      electrics.values[v.exhaustFlame] = (v.state == "firing" or v.state == "flying") and 1 or 0
      if v.hasFired then
        obj:addParticleByNodes(v.trailNodeFront, v.trailNodeBack, 0, 37, 0.01, 5) --34, 39
        obj:addParticleByNodes(v.trailNodeFront, v.trailNodeBack, 0, 43, 0.01, 5) --34, 39
        obj:addParticleByNodes(v.trailNodeFront, v.trailNodeBack, -1, 37, 0.01, 5) --34, 39
        obj:addParticleByNodes(v.trailNodeFront, v.trailNodeBack, 1, 37, 0.01, 5) --34, 39
      end
    end
  end

  activeMissiles = activeMissiles2
  activeMissileCount = #activeMissiles

  if missileCamRefNodeData then
    missileCamRefNodeData.timer = max(missileCamRefNodeData.timer - dt, 0)
    if missileCamRefNodeData.timer <= 0 and not missileCamRefNodeData.hasSetRefNodes then
      setCameraRefNodes(missileCamRefNodeData.center, missileCamRefNodeData.left, missileCamRefNodeData.back)
      missileCamRefNodeData.hasSetRefNodes = true
    end
  end
end

local function fireMissile()
  if availableMissiles[missileIndex] then
    if targetMode == "Guided" and #lockedTargets <= 0 then
      return
    end
    local missile = availableMissiles[missileIndex]
    local mode = (targetMode == "Guided" or (targetMode == "Auto-Select" and #lockedTargets > 0)) and "guided" or "nonguided"

    if mode == "guided"  then
      missile.targetID = lockedTargets[1].id
      missileLockData[missile.targetID].activeMissiles = missileLockData[missile.targetID].activeMissiles + 1
    elseif mode  == "nonguided" then
      local posRefFront = vec3(obj:getNodePosition(missile.frontCid))
      local posRefBack = vec3(obj:getNodePosition(missile.backCid))
      local directionVector = (posRefFront - posRefBack):normalized()
      missile.targetVector = directionVector
    end

    missile.state = "firing"
    missile.update = updateMissileFiring
    missile.isAlive = true
    missile.timer = 0
    missile.mode = mode
    missile.position = vec3(0,0,0)
    missile.velocity = vec3(0,0,0)
    missile.directionVector = vec3(0,0,0)
    missile.directionVectorUp = vec3(0,0,0)
    electrics.values[missile.motor] = 1
    electrics.values[missile.elevator] = 0
    electrics.values[missile.rudder] = 0
    electrics.values[missile.elevatorVect] = 0
    electrics.values[missile.rudderVect] = 0
    sounds.playSoundOnceAtNode("missileLaunchSound", (missile.thrusterNode2), 15)
    missile.hasFired = true
    table.insert(activeMissiles, missile)
    activeMissileCount = #activeMissiles
    missileIndex = missileIndex - 1
    beamstate.detachCouplers(missile.couplerTag)
    missileCamRefNodeData = {timer = 0.5, missileID = missile.id, center = missile.frontCid, left = missile.leftCid, back = missile.backCid}
  end

  gui.message(string.format("Missiles: %d", max(missileIndex, 0)), 3, "missiles")
end

local function setTargetMode(mode)
  targetMode = mode
  gui.message("Targetmode: "..targetMode, 2, "targetmode")
end

local function toggleTargetMode()
  local found = false
  local newMode = targetMode
  for _,v in pairs(targetModes) do
    if found then
      newMode = v
      found = false
      break
    elseif targetMode == v then
      found = true
    end
  end

  if found then
    newMode = targetModes[next(targetModes)]
  end

  setTargetMode(newMode)
end

local function init(jbeamData)
  --dump(jbeamData)
  activeMissiles = {}
  availableMissiles = {}
  missileLockData = {}
  lockedTargets = {}
  debugVectors = {}
  missileCamRefNodeData = nil
  activeMissileCount = 0

  electrics.values.lockingTone = 1
  electrics.values.lockTone = 1

  setTargetMode(targetMode)

  lockAngleVertical = math.rad(jbeamData.lockAngleVertical or 40)
  lockAngleHorizontal = math.rad(jbeamData.lockAngleHorizontal or 40)
  lockDistance = jbeamData.lockDistance or 5000

  local missileTargetVelocity = jbeamData.targetVelocity or 500
  local missileFuel = jbeamData.fuel or 15
  local missileExplosionNodeMass = jbeamData.explosionNodeMass or 50
  local missileAimAheadMinCoef = jbeamData.aimAheadMinCoef or 0.095
  local missileAimAheadMaxCoef = jbeamData.aimAheadMaxCoef or 0.575
  local missileAimAheadMinDistance = jbeamData.aimAheadMinDistance or 20
  local missileAimAheadMaxDistance = jbeamData.aimAheadMaxDistance or 40

  local missileFireRefs = {}
  local missileWarheadRefs = {}
  for k, node in pairs (v.data.nodes) do
    if node.fireNodeID then
      if not missileFireRefs[node.fireNodeID] then
        missileFireRefs[node.fireNodeID]= {}
      end
      table.insert(missileFireRefs[node.fireNodeID], k)
    end
    if node.warheadNodeID then
      if not missileWarheadRefs[node.warheadNodeID] then
        missileWarheadRefs[node.warheadNodeID]= {}
      end
      table.insert(missileWarheadRefs[node.warheadNodeID], k)
    end
  end

  local triggerRef = {}
  local collisionDetectorRef = {}
  for _, beam in pairs(v.data.beams) do
    if beam.weaponTriggerID then triggerRef[beam.weaponTriggerID] = beam.cid end
    if beam.collisionDetectorID then collisionDetectorRef[beam.collisionDetectorID] = beam.cid end
  end

  for k,v in pairs(jbeamData) do
    if string.sub(k, 1, 7) == "missile" then
      local missileID = v.id
      local electricsPrefix = "missile"..missileID
      local missile = {
        id = missileID,
        state = "idle",
        targetID = nil,
        nodeCid = v.centerNode,
        couplerTag = "hardpoint"..missileID,
        motor = electricsPrefix.."_motor",
        elevator = electricsPrefix.."_elevator",
        elevatorVect = electricsPrefix.."_elevator_vect",
        rudder = electricsPrefix.."_rudder",
        rudderVect = electricsPrefix.."_rudder_vect",
        exhaustFlame = electricsPrefix.."_flame",
        frontCid = v.refFront,
        backCid = v.refBack,
        leftCid = v.refLeft,
        triggerCid = triggerRef[missileID],
        collisionCid = collisionDetectorRef[missileID],
        fireCids = missileFireRefs[missileID],
        warheadCids = missileWarheadRefs[missileID],
        trailNodeFront = v.trailNodeFront,
        trailNodeBack = v.trailNodeBack,
        thrusterNode1 = v.thrusterNodeBack,
        thrusterNode2 = v.thrusterNodeFront,
        targetVelocity = v.targetVelocity or missileTargetVelocity,
        fuel = v.fuel or missileFuel,
        explosionNodeMass = v.explosionNodeMass or missileExplosionNodeMass,
        aimAheadMinCoef = v.aimAheadMinCoef or missileAimAheadMinCoef,
        aimAheadMaxCoef = v.aimAheadMaxCoef or missileAimAheadMaxCoef,
        aimAheadMinDistance = v.aimAheadMinDistance or missileAimAheadMinDistance,
        aimAheadMaxDistance = v.aimAheadMaxDistance or missileAimAheadMaxDistance,
      }
      missile.aimAheadCoefRange = missile.aimAheadMaxCoef - missile.aimAheadMinCoef
      missile.aimAheadDistanceRange = missile.aimAheadMaxDistance - missile.aimAheadMinDistance
      missile.aimAheadInvDistanceRange = 1 / missile.aimAheadDistanceRange
      electrics.values[missile.motor] = 0
      electrics.values[missile.elevator] = 0
      electrics.values[missile.elevatorVect] = 0
      electrics.values[missile.rudder] = 0
      electrics.values[missile.rudderVect] = 0
      electrics.values[missile.exhaustFlame] = 0
      table.insert(availableMissiles, missile)
    end
  end

  table.sort(availableMissiles,
    function(a,b)
      return a.id < b.id
    end)

  missileIndex = #availableMissiles

  math.randomseed(os.time())
  mapmgr.requestMap()
  resetCameraRefNodes()
end

M.init = init
M.updateGFX = updateGFX
M.update = update
M.deployWeapon = fireMissile
M.cancelWeaponCam = cancelWeaponCam
M.toggleTargetMode = toggleTargetMode
--M.debugDraw = debugDraw

return M